@@ -35,14 +35,7 @@
 #define MAX_VALUE_LEN	200
 
 struct st_engine_ctx {
-	PKCS11_CTX *pkcs11_ctx;
-	PKCS11_SLOT *slot_list;
-	unsigned int slot_count;
-#if OPENSSL_VERSION_NUMBER >= 0x10100004L
-	CRYPTO_RWLOCK *rwlock;
-#else
-	int rwlock;
-#endif
+	/* Engine configuration */
 	/*
 	 * The PIN used for login. Cache for the get_pin function.
 	 * The memory for this PIN is always owned internally,
@@ -54,8 +47,31 @@ struct st_engine_ctx {
 	int verbose;
 	char *module;
 	char *init_args;
+
+	/* Engine initialization mutex */
+#if OPENSSL_VERSION_NUMBER >= 0x10100004L
+	CRYPTO_RWLOCK *rwlock;
+#else
+	int rwlock;
+#endif
+
+	/* Current operations */
+	PKCS11_CTX *pkcs11_ctx;
+	PKCS11_SLOT *slot_list;
+	unsigned int slot_count;
 };
 
+/******************************************************************************/
+/* Utility functions                                                          */
+/******************************************************************************/
+
+static void dump_hex(FILE *stream, const char *val, const size_t len) {
+	size_t n;
+
+	for (n = 0; n < len; n++)
+		fprintf(stream, "%02x", val[n]);
+}
+
 /******************************************************************************/
 /* PIN handling                                                               */
 /******************************************************************************/
@@ -80,24 +96,22 @@ static int get_pin(ENGINE_CTX *ctx, UI_METHOD *ui_method, void *callback_data)
 	UI *ui;
 
 	/* call ui to ask for a pin */
-	ui = UI_new();
+	ui = UI_new_method(ui_method);
 	if (ui == NULL) {
 		fprintf(stderr, "UI_new failed\n");
 		return 0;
 	}
-	if (ui_method != NULL)
-		UI_set_method(ui, ui_method);
 	if (callback_data != NULL)
 		UI_add_user_data(ui, callback_data);
 
 	destroy_pin(ctx);
-	ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH * sizeof(char));
+	ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
 	if (ctx->pin == NULL)
 		return 0;
-	memset(ctx->pin, 0, MAX_PIN_LENGTH * sizeof(char));
+	memset(ctx->pin, 0, MAX_PIN_LENGTH+1);
 	ctx->pin_length = MAX_PIN_LENGTH;
 	if (!UI_add_input_string(ui, "PKCS#11 token PIN: ",
-			UI_INPUT_FLAG_DEFAULT_PWD, ctx->pin, 1, MAX_PIN_LENGTH)) {
+			UI_INPUT_FLAG_DEFAULT_PWD, ctx->pin, 4, MAX_PIN_LENGTH)) {
 		fprintf(stderr, "UI_add_input_string failed\n");
 		UI_free(ui);
 		return 0;
@@ -146,17 +160,11 @@ ENGINE_CTX *pkcs11_new()
 	return ctx;
 }
 
-int pkcs11_finish(ENGINE_CTX *ctx)
+/* Destroy the context allocated with pkcs11_new() */
+int pkcs11_destroy(ENGINE_CTX *ctx)
 {
 	if (ctx) {
-		if (ctx->slot_list) {
-			PKCS11_release_all_slots(ctx->pkcs11_ctx,
-				ctx->slot_list, ctx->slot_count);
-		}
-		if (ctx->pkcs11_ctx) {
-			PKCS11_CTX_unload(ctx->pkcs11_ctx);
-			PKCS11_CTX_free(ctx->pkcs11_ctx);
-		}
+		pkcs11_finish(ctx);
 		destroy_pin(ctx);
 		OPENSSL_free(ctx->module);
 		OPENSSL_free(ctx->init_args);
@@ -251,6 +259,29 @@ int pkcs11_init(ENGINE_CTX *ctx)
 	return 1;
 }
 
+/* Finish engine operations initialized with pkcs11_init() */
+int pkcs11_finish(ENGINE_CTX *ctx)
+{
+	if (ctx) {
+		if (ctx->slot_list) {
+			PKCS11_release_all_slots(ctx->pkcs11_ctx,
+				ctx->slot_list, ctx->slot_count);
+			ctx->slot_list = NULL;
+			ctx->slot_count = 0;
+		}
+		if (ctx->pkcs11_ctx) {
+			/* Modules cannot be unloaded in pkcs11_finish() nor
+			 * pkcs11_destroy() because of a deadlock in PKCS#11
+			 * modules that internally use OpenSSL engines.
+			 * A memory leak is better than a deadlock... */
+			/* PKCS11_CTX_unload(ctx->pkcs11_ctx); */
+			PKCS11_CTX_free(ctx->pkcs11_ctx);
+			ctx->pkcs11_ctx = NULL;
+		}
+	}
+	return 1;
+}
+
 /******************************************************************************/
 /* Certificate handling                                                       */
 /******************************************************************************/
@@ -269,8 +300,8 @@ static X509 *pkcs11_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id)
 	unsigned char cert_id[MAX_VALUE_LEN / 2];
 	size_t cert_id_len = sizeof(cert_id);
 	char *cert_label = NULL;
-	char tmp_pin[MAX_PIN_LENGTH];
-	size_t tmp_pin_len = sizeof(tmp_pin);
+	char tmp_pin[MAX_PIN_LENGTH+1];
+	size_t tmp_pin_len = MAX_PIN_LENGTH;
 	int slot_nr = -1;
 	char flags[64];
 
@@ -284,12 +315,12 @@ static X509 *pkcs11_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id)
 				tmp_pin, &tmp_pin_len, &cert_label);
 			if (n && tmp_pin_len > 0 && tmp_pin[0] != 0) {
 				destroy_pin(ctx);
-				ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH * sizeof(char));
+				ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
 				if (ctx->pin != NULL) {
+					memset(ctx->pin, 0, MAX_PIN_LENGTH+1);
 					memcpy(ctx->pin, tmp_pin, tmp_pin_len);
 					ctx->pin_length = tmp_pin_len;
 				}
-				memset(ctx->pin, 0, MAX_PIN_LENGTH * sizeof(char));
 			}
 
 			if (!n) {
@@ -314,12 +345,15 @@ static X509 *pkcs11_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id)
 		if (ctx->verbose) {
 			fprintf(stderr, "Looking in slot %d for certificate: ",
 				slot_nr);
-			if (cert_label == NULL) {
-				for (n = 0; n < cert_id_len; n++)
-					fprintf(stderr, "%02x", cert_id[n]);
-				fprintf(stderr, "\n");
-			} else
-				fprintf(stderr, "label: %s\n", cert_label);
+			if (cert_id_len != 0) {
+				fprintf(stderr, "id=");
+				dump_hex(stderr, cert_id, cert_id_len);
+			}
+			if (cert_id_len != 0 && cert_label != NULL)
+				fprintf(stderr, " ");
+			if (cert_label != NULL)
+				fprintf(stderr, "label=%s", cert_label);
+			fprintf(stderr, "\n");
 		}
 	}
 
@@ -429,25 +463,22 @@ static X509 *pkcs11_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id)
 		for (n = 0; n < cert_count; n++) {
 			PKCS11_CERT *k = certs + n;
 
-			if (cert_label == NULL) {
-				if (cert_id_len != 0 && k->id_len == cert_id_len &&
-						memcmp(k->id, cert_id, cert_id_len) == 0)
-					selected_cert = k;
-			} else {
-				if (strcmp(k->label, cert_label) == 0)
-					selected_cert = k;
-			}
+			if (cert_label != NULL && strcmp(k->label, cert_label) == 0)
+				selected_cert = k;
+			if (cert_id_len != 0 && k->id_len == cert_id_len &&
+					memcmp(k->id, cert_id, cert_id_len) == 0)
+				selected_cert = k;
 		}
 	} else {
 		selected_cert = certs; /* Use the first certificate */
 	}
 
-	if (selected_cert == NULL) {
+	if (selected_cert != NULL) {
+		x509 = X509_dup(selected_cert->x509);
+	} else {
 		fprintf(stderr, "Certificate not found.\n");
-		return NULL;
+		x509 = NULL;
 	}
-
-	x509 = X509_dup(selected_cert->x509);
 	if (cert_label != NULL)
 		OPENSSL_free(cert_label);
 	return x509;
@@ -495,13 +526,13 @@ static int pkcs11_login(ENGINE_CTX *ctx, PKCS11_SLOT *slot, PKCS11_TOKEN *tok,
 			 * assigned (i.e, cached by get_pin) */
 			destroy_pin(ctx);
 		} else if (ctx->pin == NULL) {
-			ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH * sizeof(char));
+			ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
 			ctx->pin_length = MAX_PIN_LENGTH;
 			if (ctx->pin == NULL) {
 				fprintf(stderr, "Could not allocate memory for PIN");
 				return 0;
 			}
-			memset(ctx->pin, 0, MAX_PIN_LENGTH * sizeof(char));
+			memset(ctx->pin, 0, MAX_PIN_LENGTH+1);
 			if (!get_pin(ctx, ui_method, callback_data)) {
 				destroy_pin(ctx);
 				fprintf(stderr, "No pin code was entered");
@@ -534,7 +565,7 @@ static int pkcs11_login(ENGINE_CTX *ctx, PKCS11_SLOT *slot, PKCS11_TOKEN *tok,
 }
 
 static EVP_PKEY *pkcs11_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
-		UI_METHOD * ui_method, void *callback_data, int isPrivate)
+		UI_METHOD *ui_method, void *callback_data, int isPrivate)
 {
 	PKCS11_SLOT *slot;
 	PKCS11_SLOT *found_slot = NULL;
@@ -547,9 +578,10 @@ static EVP_PKEY *pkcs11_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 	size_t key_id_len = sizeof(key_id);
 	char *key_label = NULL;
 	int slot_nr = -1;
-	char tmp_pin[MAX_PIN_LENGTH];
-	size_t tmp_pin_len = sizeof(tmp_pin);
+	char tmp_pin[MAX_PIN_LENGTH+1];
+	size_t tmp_pin_len = MAX_PIN_LENGTH;
 	char flags[64];
+	int already_logged_in = 0;
 
 	if (pkcs11_init_libp11(ctx)) /* Delayed libp11 initialization */
 		return NULL;
@@ -566,9 +598,9 @@ static EVP_PKEY *pkcs11_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 
 			if (n && tmp_pin_len > 0 && tmp_pin[0] != 0) {
 				destroy_pin(ctx);
-				ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH * sizeof(char));
+				ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
 				if (ctx->pin != NULL) {
-					memset(ctx->pin, 0, MAX_PIN_LENGTH * sizeof(char));
+					memset(ctx->pin, 0, MAX_PIN_LENGTH+1);
 					memcpy(ctx->pin, tmp_pin, tmp_pin_len);
 					ctx->pin_length = tmp_pin_len;
 				}
@@ -596,12 +628,15 @@ static EVP_PKEY *pkcs11_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 		if (ctx->verbose) {
 			fprintf(stderr, "Looking in slot %d for key: ",
 				slot_nr);
-			if (key_label == NULL) {
-				for (n = 0; n < key_id_len; n++)
-					fprintf(stderr, "%02x", key_id[n]);
-				fprintf(stderr, "\n");
-			} else
-				fprintf(stderr, "label: %s\n", key_label);
+			if (key_id_len != 0) {
+				fprintf(stderr, "id=");
+				dump_hex(stderr, key_id, key_id_len);
+			}
+			if (key_id_len != 0 && key_label != NULL)
+				fprintf(stderr, " ");
+			if (key_label != NULL)
+				fprintf(stderr, "label=%s", key_label);
+			fprintf(stderr, "\n");
 		}
 	}
 
@@ -706,7 +741,9 @@ static EVP_PKEY *pkcs11_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 			PKCS11_CERT *c = certs + n;
 			char *dn = NULL;
 
-			fprintf(stderr, "  %2u    %s", n + 1, c->label);
+			fprintf(stderr, "  %2u    id=", n + 1);
+			dump_hex(stderr, c->id, c->id_len);
+			fprintf(stderr, " label=%s", c->label);
 			if (c->x509)
 				dn = X509_NAME_oneline(X509_get_subject_name(c->x509), NULL, 0);
 			if (dn) {
@@ -718,12 +755,16 @@ static EVP_PKEY *pkcs11_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 	}
 
 	if (isPrivate) {
+		/* Check if already logged in to avoid resetting state */
+		if (PKCS11_is_logged_in(slot, 0, &already_logged_in) != 0) {
+			fprintf(stderr, "Unable to check if already logged in\n");
+			return NULL;
+		}
 		/* Perform login to the token if required */
-		if (!pkcs11_login(ctx, slot, tok, ui_method, callback_data)) {
+		if (!already_logged_in && !pkcs11_login(ctx, slot, tok, ui_method, callback_data)) {
 			fprintf(stderr, "login to token failed, returning NULL...\n");
 			return NULL;
 		}
-
 		/* Make sure there is at least one private key on the token */
 		if (PKCS11_enumerate_keys(tok, &keys, &key_count)) {
 			fprintf(stderr, "Unable to enumerate private keys\n");
@@ -752,40 +793,37 @@ static EVP_PKEY *pkcs11_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 			PKCS11_KEY *k = keys + n;
 
 			if (ctx->verbose) {
-				fprintf(stderr, "  %2u %c%c %s\n", n + 1,
+				fprintf(stderr, "  %2u %c%c id=", n + 1,
 					k->isPrivate ? 'P' : ' ',
-					k->needLogin ? 'L' : ' ', k->label);
-			}
-			if (key_label == NULL) {
-				if (key_id_len != 0 && k->id_len == key_id_len
-						&& memcmp(k->id, key_id, key_id_len) == 0) {
-					selected_key = k;
-				}
-			} else {
-				if (strcmp(k->label, key_label) == 0) {
-					selected_key = k;
-				}
+					k->needLogin ? 'L' : ' ');
+				dump_hex(stderr, k->id, k->id_len);
+				fprintf(stderr, " label=%s\n", k->label);
 			}
+			if (key_label != NULL && strcmp(k->label, key_label) == 0)
+				selected_key = k;
+			if (key_id_len != 0 && k->id_len == key_id_len
+					&& memcmp(k->id, key_id, key_id_len) == 0)
+				selected_key = k;
 		}
 	} else {
 		selected_key = keys; /* Use the first key */
 	}
 
-	if (selected_key == NULL) {
+	if (selected_key != NULL) {
+		pk = isPrivate ?
+			PKCS11_get_private_key(selected_key) :
+			PKCS11_get_public_key(selected_key);
+	} else {
 		fprintf(stderr, "Key not found.\n");
-		return NULL;
+		pk = NULL;
 	}
-
-	pk = isPrivate ?
-		PKCS11_get_private_key(selected_key) :
-		PKCS11_get_public_key(selected_key);
 	if (key_label != NULL)
 		OPENSSL_free(key_label);
 	return pk;
 }
 
 EVP_PKEY *pkcs11_load_public_key(ENGINE_CTX *ctx, const char *s_key_id,
-		UI_METHOD * ui_method, void *callback_data)
+		UI_METHOD *ui_method, void *callback_data)
 {
 	EVP_PKEY *pk;
 
@@ -798,7 +836,7 @@ EVP_PKEY *pkcs11_load_public_key(ENGINE_CTX *ctx, const char *s_key_id,
 }
 
 EVP_PKEY *pkcs11_load_private_key(ENGINE_CTX *ctx, const char *s_key_id,
-		UI_METHOD * ui_method, void *callback_data)
+		UI_METHOD *ui_method, void *callback_data)
 {
 	EVP_PKEY *pk;
 
