@@ -36,9 +36,12 @@ static int rsa_ex_index = 0;
 static RSA *pkcs11_rsa(PKCS11_KEY *key)
 {
 	EVP_PKEY *evp_key = pkcs11_get_key(key, key->isPrivate);
+	RSA *rsa;
 	if (evp_key == NULL)
 		return NULL;
-	return EVP_PKEY_get0_RSA(evp_key);
+	rsa = EVP_PKEY_get0_RSA(evp_key);
+	EVP_PKEY_free(evp_key);
+	return rsa;
 }
 
 /* PKCS#1 v1.5 RSA signature */
@@ -96,7 +99,7 @@ int pkcs11_private_encrypt(int flen,
 	/* Try signing first, as applications are more likely to use it */
 	rv = CRYPTOKI_call(ctx,
 		C_SignInit(spriv->session, &mechanism, kpriv->object));
-	if (!rv)
+	if (rv == CKR_USER_NOT_LOGGED_IN)
 		rv = pkcs11_authenticate(key);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
@@ -105,7 +108,7 @@ int pkcs11_private_encrypt(int flen,
 		/* OpenSSL may use it for encryption rather than signing */
 		rv = CRYPTOKI_call(ctx,
 			C_EncryptInit(spriv->session, &mechanism, kpriv->object));
-		if (!rv)
+		if (rv == CKR_USER_NOT_LOGGED_IN)
 			rv = pkcs11_authenticate(key);
 		if (!rv)
 			rv = CRYPTOKI_call(ctx,
@@ -139,7 +142,7 @@ int pkcs11_private_decrypt(int flen, const unsigned char *from, unsigned char *t
 	CRYPTO_THREAD_write_lock(PRIVSLOT(slot)->rwlock);
 	rv = CRYPTOKI_call(ctx,
 		C_DecryptInit(spriv->session, &mechanism, kpriv->object));
-	if (!rv)
+	if (rv == CKR_USER_NOT_LOGGED_IN)
 		rv = pkcs11_authenticate(key);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
@@ -179,23 +182,20 @@ static RSA *pkcs11_get_rsa(PKCS11_KEY *key)
 	RSA *rsa;
 	PKCS11_KEY *keys = NULL;
 	unsigned int i, count = 0;
+	BIGNUM *rsa_n=NULL, *rsa_e=NULL;
 
 	rsa = RSA_new();
 	if (rsa == NULL)
 		return NULL;
 
 	/* Retrieve the modulus and the public exponent */
-	if (key_getattr_bn(key, CKA_MODULUS, &rsa->n) ||
-			key_getattr_bn(key, CKA_PUBLIC_EXPONENT, &rsa->e)) {
-		RSA_free(rsa);
-		return NULL;
-	}
-	if(!BN_is_zero(rsa->e)) /* The public exponent was retrieved */
-		return rsa;
-	BN_clear_free(rsa->e);
-	/* In case someone modifies this function to execute RSA_free()
-	 * before a valid BN value is assigned to rsa->e */
-	rsa->e = NULL;
+	if (key_getattr_bn(key, CKA_MODULUS, &rsa_n) ||
+			key_getattr_bn(key, CKA_PUBLIC_EXPONENT, &rsa_e))
+		goto failure;
+	if (!BN_is_zero(rsa_e)) /* The public exponent was retrieved */
+		goto success;
+	BN_clear_free(rsa_e);
+	rsa_e = NULL;
 
 	/* The public exponent was not found in the private key:
 	 * retrieve it from the corresponding public key */
@@ -205,11 +205,11 @@ static RSA *pkcs11_get_rsa(PKCS11_KEY *key)
 
 			if (key_getattr_bn(&keys[i], CKA_MODULUS, &pubmod))
 				continue; /* Failed to retrieve the modulus */
-			if (BN_cmp(rsa->n, pubmod) == 0) { /* The key was found */
+			if (BN_cmp(rsa_n, pubmod) == 0) { /* The key was found */
 				BN_clear_free(pubmod);
-				if (key_getattr_bn(&keys[i], CKA_PUBLIC_EXPONENT, &rsa->e))
+				if (key_getattr_bn(&keys[i], CKA_PUBLIC_EXPONENT, &rsa_e))
 					continue; /* Failed to retrieve the public exponent */
-				return rsa;
+				goto success;
 			} else {
 				BN_clear_free(pubmod);
 			}
@@ -217,12 +217,22 @@ static RSA *pkcs11_get_rsa(PKCS11_KEY *key)
 	}
 
 	/* Last resort: use the most common default */
-	rsa->e = BN_new();
-	if(rsa->e && BN_set_word(rsa->e, RSA_F4))
-		return rsa;
+	rsa_e = BN_new();
+	if (rsa_e && BN_set_word(rsa_e, RSA_F4))
+		goto success;
 
+failure:
 	RSA_free(rsa);
 	return NULL;
+
+success:
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L
+		RSA_set0_key(rsa, rsa_n, rsa_e, NULL);
+#else
+		rsa->n=rsa_n;
+		rsa->e=rsa_e;
+#endif
+	return rsa;
 }
 
 /*
@@ -261,9 +271,16 @@ static EVP_PKEY *pkcs11_get_evp_key_rsa(PKCS11_KEY *key)
 int pkcs11_get_key_modulus(PKCS11_KEY *key, BIGNUM **bn)
 {
 	RSA *rsa = pkcs11_rsa(key);
+	const BIGNUM *rsa_n;
+
 	if (rsa == NULL)
 		return 0;
-	*bn = BN_dup(rsa->n);
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L
+	RSA_get0_key(rsa, &rsa_n, NULL, NULL);
+#else
+	rsa_n=rsa->n;
+#endif
+	*bn = BN_dup(rsa_n);
 	return *bn == NULL ? 0 : 1;
 }
 
@@ -271,9 +288,16 @@ int pkcs11_get_key_modulus(PKCS11_KEY *key, BIGNUM **bn)
 int pkcs11_get_key_exponent(PKCS11_KEY *key, BIGNUM **bn)
 {
 	RSA *rsa = pkcs11_rsa(key);
+	const BIGNUM *rsa_e;
+
 	if (rsa == NULL)
 		return 0;
-	*bn = BN_dup(rsa->e);
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L
+	RSA_get0_key(rsa, NULL, &rsa_e, NULL);
+#else
+	rsa_e=rsa->e;
+#endif
+	*bn = BN_dup(rsa_e);
 	return *bn == NULL ? 0 : 1;
 }
 
@@ -286,13 +310,33 @@ int pkcs11_get_key_size(PKCS11_KEY *key)
 	return RSA_size(rsa);
 }
 
+#if OPENSSL_VERSION_NUMBER < 0x10100005L
+
+int (*RSA_meth_get_priv_enc(const RSA_METHOD *meth))
+		(int flen, const unsigned char *from,
+			unsigned char *to, RSA *rsa, int padding)
+{
+    return meth->rsa_priv_enc;
+}
+
+int (*RSA_meth_get_priv_dec(const RSA_METHOD *meth))
+		(int flen, const unsigned char *from,
+			unsigned char *to, RSA *rsa, int padding)
+{
+    return meth->rsa_priv_dec;
+}
+
+#endif
+
 static int pkcs11_rsa_priv_dec_method(int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding)
 {
 	PKCS11_KEY *key = RSA_get_ex_data(rsa, rsa_ex_index);
+	int (*priv_dec) (int flen, const unsigned char *from,
+		unsigned char *to, RSA *rsa, int padding);
 	if (key == NULL) {
-		PKCS11err(PKCS11_F_PKCS11_RSA_DECRYPT, PKCS11_ALIEN_KEY);
-		return -1;
+		priv_dec = RSA_meth_get_priv_dec(RSA_get_default_method());
+		return priv_dec(flen, from, to, rsa, padding);
 	}
 	return PKCS11_private_decrypt(flen, from, to, key, padding);
 }
@@ -301,9 +345,11 @@ static int pkcs11_rsa_priv_enc_method(int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding)
 {
 	PKCS11_KEY *key = RSA_get_ex_data(rsa, rsa_ex_index);
+	int (*priv_enc) (int flen, const unsigned char *from,
+		unsigned char *to, RSA *rsa, int padding);
 	if (key == NULL) {
-		PKCS11err(PKCS11_F_PKCS11_RSA_ENCRYPT, PKCS11_ALIEN_KEY);
-		return -1;
+		priv_enc = RSA_meth_get_priv_enc(RSA_get_default_method());
+		return priv_enc(flen, from, to, rsa, padding);
 	}
 	return PKCS11_private_encrypt(flen, from, to, key, padding);
 }
@@ -336,6 +382,62 @@ static void free_rsa_ex_index()
 #endif
 }
 
+#if OPENSSL_VERSION_NUMBER < 0x10100005L
+
+static RSA_METHOD *RSA_meth_dup(const RSA_METHOD *meth)
+{
+	RSA_METHOD *ret = OPENSSL_malloc(sizeof(RSA_METHOD));
+	if (ret == NULL)
+		return NULL;
+	memcpy(ret, meth, sizeof(RSA_METHOD));
+	ret->name = OPENSSL_strdup(meth->name);
+	if (ret->name == NULL) {
+		OPENSSL_free(ret);
+		return NULL;
+	}
+	return ret;
+}
+
+static int RSA_meth_set1_name(RSA_METHOD *meth, const char *name)
+{
+	char *tmp = OPENSSL_strdup(name);
+	if (tmp == NULL)
+		return 0;
+	OPENSSL_free((char *)meth->name);
+	meth->name = tmp;
+	return 1;
+}
+
+static int RSA_meth_set_flags(RSA_METHOD *meth, int flags)
+{
+	meth->flags = flags;
+	return 1;
+}
+
+static int RSA_meth_set_priv_enc(RSA_METHOD *meth,
+		int (*priv_enc) (int flen, const unsigned char *from,
+		unsigned char *to, RSA *rsa, int padding))
+{
+	meth->rsa_priv_enc = priv_enc;
+	return 1;
+}
+
+static int RSA_meth_set_priv_dec(RSA_METHOD *meth,
+		int (*priv_dec) (int flen, const unsigned char *from,
+		unsigned char *to, RSA *rsa, int padding))
+{
+	meth->rsa_priv_dec = priv_dec;
+	return 1;
+}
+
+static int RSA_meth_set_finish(RSA_METHOD *meth, int (*finish)(RSA *rsa))
+{
+	meth->finish = finish;
+	return 1;
+}
+
+#endif
+
 /*
  * Overload the default OpenSSL methods for RSA
  */
@@ -345,13 +447,14 @@ RSA_METHOD *PKCS11_get_rsa_method(void)
 
 	if (ops == NULL) {
 		alloc_rsa_ex_index();
-		ops = OPENSSL_malloc(sizeof(RSA_METHOD));
+		ops = RSA_meth_dup(RSA_get_default_method());
 		if (ops == NULL)
 			return NULL;
-		memcpy(ops, RSA_get_default_method(), sizeof(RSA_METHOD));
-		ops->rsa_priv_enc = pkcs11_rsa_priv_enc_method;
-		ops->rsa_priv_dec = pkcs11_rsa_priv_dec_method;
-		ops->finish = pkcs11_rsa_free_method;
+		RSA_meth_set1_name(ops, "libp11 RSA method");
+		RSA_meth_set_flags(ops, 0);
+		RSA_meth_set_priv_enc(ops, pkcs11_rsa_priv_enc_method);
+		RSA_meth_set_priv_dec(ops, pkcs11_rsa_priv_dec_method);
+		RSA_meth_set_finish(ops, pkcs11_rsa_free_method);
 	}
 	return ops;
 }
